import ssl

from tests.http_test_servers import TLSTestServer
from tests.utils import get_remote_certificate


def test_dummy_tls_server(tmp_path):
    """
    Tests that the test server works expected.

    That TLS server doesn't have the SNI extension set-up thus
    it should always return one and only one certificate.
    """

    fqn = "domain.local"
    srv = TLSTestServer(tmp_path, cert_identities=[fqn], cert_options={})

    with srv as sock_addr:
        ssl_ctx = ssl.create_default_context()
        ssl_ctx.load_verify_locations(srv.ca_bundle_path)

        # Get the certificate a domain that is associated with the server's
        # TLS certificate.
        cert1 = get_remote_certificate(ssl_ctx, sock_addr, server_hostname=fqn)

        # Retrieve the server's certificate returned when an unknown domain
        # is being connected onto.
        ssl_ctx.check_hostname = False  # The hostname will mismatch thus ignore it
        cert2 = get_remote_certificate(
            ssl_ctx, sock_addr, server_hostname="invalid.local"
        )

    # domain.local and invalid.local should have returned the same certificate.
    assert cert1 == cert2



from unittest import mock

import pytest

from requests_hardened import Manager, Config

DEFAULT_TIMEOUT = (1, 2)

NeverRedirectManager = Manager(
    Config(
        never_redirect=True,
        # irrelevant:
        ip_filter_enable=False,
        ip_filter_allow_loopback_ips=True,
        default_timeout=(1, 1),
        user_agent_override="user-agent",
    )
)

AllowRedirectManager = NeverRedirectManager.clone()
AllowRedirectManager.config.never_redirect = False


@pytest.mark.parametrize(
    "_name, kwargs",
    [
        ("allow_redirects not passed -> should set to False", {}),
        (
            "allow_redirects=True passed -> should set to False",
            {"allow_redirects": True},
        ),
        (
            "allow_redirects=False passed -> should remain False",
            {"allow_redirects": False},
        ),
    ],
)
@mock.patch("requests.sessions.Session.send")
def test_never_allows_redirects(mocked_send: mock.MagicMock, _name, kwargs):
    NeverRedirectManager.send_request("GET", "https://example.com", **kwargs)
    mocked_send.assert_called_once_with(
        mock.ANY,
        timeout=NeverRedirectManager.config.default_timeout,
        allow_redirects=False,
        proxies={},
        stream=False,
        verify=True,
        cert=None,
    )


@pytest.mark.parametrize(
    "_name, kwargs, expected_allow_redirects",
    [
        ("allow_redirects not passed -> should set to True", {}, True),
        (
            "allow_redirects=False passed -> should remain False",
            {"allow_redirects": False},
            False,
        ),
        (
            "allow_redirects=True passed -> should remain True",
            {"allow_redirects": True},
True,
        ),
    ],
)
@mock.patch("requests.sessions.Session.send")
def test_allows_redirects(mocked_send: mock.MagicMock, _name, kwargs, expected_allow_redirects):
    AllowRedirectManager.send_request("GET", "https://example.com", **kwargs)
    mocked_send.assert_called_once_with(
        mock.ANY,
        timeout=NeverRedirectManager.config.default_timeout,
        allow_redirects=expected_allow_redirects,
        proxies={},
        stream=False,
        verify=True,
        cert=None,
    )



import functools
import ipaddress
import socket
import sys
from socket import AddressFamily, SocketKind
from unittest import mock

import pytest
from requests import PreparedRequest
from requests.exceptions import SSLError
from urllib3.exceptions import MaxRetryError
from urllib3.util.ssl_match_hostname import CertificateError

from requests_hardened import Manager, Config
from .http_test_servers import (
    InsecureHTTPTestServer,
    TLSTestServer,
    SNITLSHTTPTestServer,
)
from .utils import mock_getaddrinfo, disable_sni_support

from requests_hardened.ip_filter import InvalidIPAddress, get_ip_address

# HTTP server can take some time to connect against for some machines (usually up to 1s).
# It requires a fairly generous timeout without taking too long that the test fails.
SOCKET_TIMEOUT = (4, 0.1)

TEST_TIMEOUT_SECS = 5


SSRFFilter = Manager(
    Config(
        default_timeout=SOCKET_TIMEOUT,
        never_redirect=False,
        ip_filter_enable=True,
        ip_filter_allow_loopback_ips=False,
        user_agent_override="user-agent",
    )
)

SSRFFilterAllowLocalHost = SSRFFilter.clone()
SSRFFilterAllowLocalHost.config.ip_filter_allow_loopback_ips = True


@pytest.mark.parametrize(
    "ip_addr",
    [
        "0.0.0.0",
        "127.0.0.1",
        "10.0.0.1",
        "192.168.0.1",
        "172.16.0.1",
        "192.0.2.0",  # broadcast address
        "::",  # 0.0.0.0
        "::1",  # 127.0.0.1
        "::ffff:7f00:1",  # 127.0.0.1
        "::ffff:a00:1",  # 10.0.0.1
        "::ffff:c0a8:1",  # 192.168.0.1
        "::ffff:ac10:1",  # 172.16.0.1
        "::ffff:192.168.2.1",
        "::ffff:192.0.2.0",  # broadcast address
        "2001:0db8:85a3:0000:0000:8a2e:0370:7334",
    ],
)
@pytest.mark.fake_resolver(enabled=False)
@mock.patch("requests.sessions.Session.send")
def test_blocks_private_ranges(mocked_send: mock.MagicMock, ip_addr: str):
    """Ensure private blocks are rejected."""
    with mock_getaddrinfo(ip_addr):
        with pytest.raises(InvalidIPAddress):
            SSRFFilter.send_request("GET", "https://test.local")
        mocked_send.assert_not_called()


@pytest.mark.parametrize(
    "ip_addr, expected_url",
    [
        ("128.99.0.0", "https://128.99.0.0:443/"),
        ("8.0.0.0", "https://8.0.0.0:443/"),
        ("::192.0.5.0", "https://[::c000:500]:443/"),  # Not IPv4
        ("::ffff:8063:0", "https://128.99.0.0:443/"),  # IPv6 mapped to IPv4
        ("::ffff:800:0", "https://8.0.0.0:443/"),  # IPv6 mapped to IPv4
        (
            "2004:0db8:1234:5678::abcd:ef01",
            "https://[2004:db8:1234:5678::abcd:ef01]:443/",
        ),
    ],
)
@pytest.mark.fake_resolver(enabled=False)
@mock.patch("requests.sessions.Session.send")
def test_allows_public_ranges(
    mocked_send: mock.MagicMock, ip_addr: str, expected_url: str
):
    """Ensure public blocks are allowed."""

    with mock_getaddrinfo(ip_addr):
        SSRFFilter.send_request("GET", "https://test.local")
        mocked_send.assert_called_once()

        [prepared_request], kwargs = mocked_send.call_args
        assert isinstance(prepared_request, PreparedRequest)
        assert prepared_request.method == "GET"
        assert prepared_request.url == expected_url
        assert prepared_request.headers.get("Host") == "test.local"


@pytest.mark.parametrize(
    "input_url, resolves_to, expected_output_url, expected_http_host_header",
    [
        (
            # Test handling of domains returning IP addresses version 6.
            # Ensures the IP is surrounded by brackets.
            "https://example.com",
            "2004:0db8:1234:5678::abcd:ef01",
            "https://[2004:db8:1234:5678::abcd:ef01]:443/",
            "example.com",
        ),
        (
            # Ensure IPv6 passed directly into a URL is handled properly:
            # - DNS resolver should be invoked using the IPv6 address without brackets;
            # - HTTP host header should be the IPv6 address (no brackets);
            # - HTTP URL should contain brackets.
            "https://[2004:0db8:1234:5678::abcd:ef01]",
            "2004:0db8:1234:5678::abcd:ef01",
            "https://[2004:db8:1234:5678::abcd:ef01]:443/",
            "2004:0db8:1234:5678::abcd:ef01",
        ),
        (
            # Ensure ports are handled properly
            "https://[2004:0db8:1234:5678::abcd:ef01]:444",
            "2004:0db8:1234:5678::abcd:ef01",
            "https://[2004:db8:1234:5678::abcd:ef01]:444/",
            "2004:0db8:1234:5678::abcd:ef01",
        ),
        (
            # Ensure ports are handled properly
            "http://[2004:0db8:1234:5678::abcd:ef01]:444",
            "2004:0db8:1234:5678::abcd:ef01",
            "http://[2004:db8:1234:5678::abcd:ef01]:444/",
            "2004:0db8:1234:5678::abcd:ef01",
        ),
        (
            # Ensure ports are handled properly
            "http://127.0.0.1:444",
            "127.0.0.1",
            "http://127.0.0.1:444/",
            "127.0.0.1",
        ),
        (
            # Ensure ports are handled properly
            "http://example.com:444",
            "127.0.0.1",
            "http://127.0.0.1:444/",
            "example.com",
        ),
        (
            # Ensure unicode URLs are encoded to IDNA
            "https://wroc\u0142aw.test",
            "127.0.0.1",
            "https://127.0.0.1:443/",
            "xn--wrocaw-6db.test",
        ),
    ],
)
@mock.patch("requests.sessions.Session.send")
def test_url_handling(
    mocked_send: mock.MagicMock,
    input_url: str,
    resolves_to: str,
    expected_output_url: str,
    expected_http_host_header: str,
):
    manager = SSRFFilter.clone()
    manager.config.ip_filter_allow_loopback_ips = True
    with mock_getaddrinfo(resolves_to):
        manager.send_request("GET", input_url)

    mocked_send.assert_called_once()

    [prepared_request], kwargs = mocked_send.call_args
    assert isinstance(prepared_request, PreparedRequest)
    assert prepared_request.method == "GET"
    assert prepared_request.url == expected_output_url
    assert prepared_request.headers.get("Host") == expected_http_host_header


@pytest.mark.parametrize(
    "resolver_result, expected_result",
    [
        # (af, socktype, proto, _canonname, socket_address)
        [
            (
                AddressFamily.AF_INET6,
                SocketKind.SOCK_STREAM,
                6,
                "",
                ("::ffff:8.8.8.0", 443, 0, 0),
            ),
            (ipaddress.IPv4Address("8.8.8.0"), 443),
        ],
        [
            (AddressFamily.AF_INET, SocketKind.SOCK_STREAM, 6, "", ("8.8.4.0", 443)),
            (ipaddress.IPv4Address("8.8.4.0"), 443),
        ],
    ],
)
@pytest.mark.fake_resolver(enabled=False)
def test_socket_family_handling(resolver_result, expected_result):
    """
    Verify the handling of different socket families: AF_INET, AF_INET6,
    AF_NETLINK, AF_TIPC, AF_UNIX. We only expect to handle AF_INET and AF_INET6.

    Length difference of socket address between INET and INET6 should be handled
    successfully.
    """

    with mock.patch.object(socket, "getaddrinfo", return_value=[resolver_result]):
        result = get_ip_address("test.local", 443, allow_loopback=False)

    assert result == expected_result


@pytest.mark.skipif(sys.platform.startswith("win"), reason="Unix-like specific test")
@pytest.mark.fake_resolver(enabled=False)
def test_rejects_non_inet_socket_family():
    """
    Verifies if the socket family is not AF_INET or AF_INET6, then it
    raises `ValueError`.
    """

    with mock.patch.object(
        socket,
        "getaddrinfo",
        return_value=[
            # AF_UNIX is a non-IP based network protocol thus the
            # getaddrinfo return value is invalid, but it allows to ensure our behavior.
            (AddressFamily.AF_UNIX, SocketKind.SOCK_STREAM, 6, "", ())
        ],
    ):
        with pytest.raises(ValueError) as exc:
            get_ip_address("test.local", 443, allow_loopback=False)

    assert exc.value.args == (
        "Only AF_INET and AF_INET6 socket address families are supported",
    )


@pytest.mark.timeout(TEST_TIMEOUT_SECS)
@pytest.mark.fake_resolver(enabled=True)
def test_insecure_http_supported():
    """
    Ensure we are able to connect to targets that are using insecure HTTP.
    """
    srv = InsecureHTTPTestServer()
    with srv as [_srv_addr, srv_port]:
        response = SSRFFilterAllowLocalHost.send_request(
            "GET", f"http://test.local:{srv_port}"
        )
        assert response.status_code == 200


@pytest.mark.timeout(TEST_TIMEOUT_SECS)
@pytest.mark.fake_resolver(enabled=True)
def test_tls_without_SNIs_supported(tmp_path):
    """
    Ensure we are able to connect successfully to a server that doesn't
    have a SNI callback set-up.
    """
    srv = TLSTestServer(
        tmp_path,
        cert_identities=["test1.local", "test2.local"],
    )
    with srv as [_srv_addr, srv_port]:
        do_request = functools.partial(
            SSRFFilterAllowLocalHost.send_request,
            "GET",
            f"https://test2.local:{srv_port}",
            verify=srv.ca_bundle_path,
        )

        # This should work both when SNI support is enabled and disabled.
        assert do_request().status_code == 200

        # Ensure this test is not testing SNI: when disabling SNI support client-side,
        # this the HTTP request should still succeed.
        with disable_sni_support():
            assert do_request().status_code == 200


@pytest.mark.timeout(TEST_TIMEOUT_SECS)
@pytest.mark.fake_resolver(enabled=True)
def test_tls_with_SNIs_supported(tmp_path):
    """
    Ensure we are able to connect successfully to a server that has a
    SNI callback set-up.
    """
    srv = SNITLSHTTPTestServer(
        tmp_path, cert_identities=["localhost"], additional_identities=["sni.local"]
    )
    with srv as [_srv_addr, srv_port]:
        do_request = functools.partial(
            SSRFFilterAllowLocalHost.send_request,
            "GET",
            f"https://sni.local:{srv_port}",
            verify=srv.ca_bundle_path,
        )

        # This should only work when SNI support is enabled.
        assert do_request().status_code == 200

        # Ensure when SNI support is disabled client-side, then
        # it no longer works, which allows to be sure the test is indeed testing SNI.
        with pytest.raises(SSLError) as exc_info:
            with disable_sni_support():
                assert do_request().status_code == 200

    exc = exc_info.value.args[0]
    assert isinstance(exc, MaxRetryError)
    original_exception = exc.reason.args[0]
    assert isinstance(original_exception, CertificateError)
    assert (
        original_exception.args[0] == "hostname 'sni.local' doesn't match 'localhost'"
    )



# noqa: D205,D212,D400,D415,D404
"""
This file is a copy of a yet to be released `HostHeaderSSLAdapter` class
with SNI support: https://github.com/requests/toolbelt/blob/c5ac5f3b5c7ed599703e0deb4fd31ce7998711ed/requests_toolbelt/adapters/host_header_ssl.py

Copyright 2014 Ian Cordasco, Cory Benfield

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       https://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

from requests.adapters import HTTPAdapter


class HostHeaderSSLAdapter(HTTPAdapter):
    # noqa: D205,D400,D415
    """An HTTPS Adapter for Python Requests that sets the hostname for certificate
    verification based on the Host header.

    This allows requesting the IP address directly via HTTPS without getting
    a "hostname doesn't match" exception.

    Example usage:

        >>> s.mount('https://', HostHeaderSSLAdapter())
        >>> s.get("https://93.184.216.34", headers={"Host": "example.org"})

    """

    def send(self, request, **kwargs):
        # HTTP headers are case-insensitive (RFC 7230)
        host_header = None
        for header in request.headers:
            if header.lower() == "host":
                host_header = request.headers[header]
                break

        connection_pool_kwargs = self.poolmanager.connection_pool_kw

        if host_header:
            connection_pool_kwargs["assert_hostname"] = host_header
            connection_pool_kwargs["server_hostname"] = host_header
        elif "assert_hostname" in connection_pool_kwargs:
            # connection pool arguments from a previous request may have been left
            connection_pool_kwargs.pop("assert_hostname", None)
            connection_pool_kwargs.pop("server_hostname", None)

        return super().send(request, **kwargs)



from typing import Union, cast, Any

import requests
from requests import Request, PreparedRequest, Response

from requests_hardened.config import Config
from requests_hardened.host_header_adapter import HostHeaderSSLAdapter
from requests_hardened.ip_filter import filter_request

T_TIMEOUT = Union[int, float]


class HTTPSession(requests.Session):
    def __init__(self, config: Config, **kwargs):
        super().__init__(**kwargs)
        self.mount("https://", HostHeaderSSLAdapter())
        self._config = config

    def send(self, request: PreparedRequest, **kwargs: Any) -> Response:
        allow_redirects = kwargs.setdefault(
            "allow_redirects", not self._config.never_redirect
        )
        if allow_redirects and self._config.never_redirect:
            kwargs["allow_redirects"] = False

        timeout = kwargs.setdefault("timeout", self._config.default_timeout)
        if not timeout:
            kwargs["timeout"] = self._config.default_timeout
        return super().send(request, **kwargs)

    def prepare_request(self, request: Request) -> PreparedRequest:
        url = request.url
        if self._config.ip_filter_enable is True:
            headers = request.headers or {}
            # Cast potentially immutable header list to `dict`
            if not isinstance(headers, dict):
                headers = cast(dict, dict(**headers))
            # Cast `bytes` to `str`
            if isinstance(url, bytes):
                url = url.decode()

            url = filter_request(
                url,
                headers=headers,
                allow_loopback=self._config.ip_filter_allow_loopback_ips,
            )
            request.url = url
            request.headers = headers
        if self._config.user_agent_override is not None:
            request.headers.update({"User-Agent": self._config.user_agent_override})
        return super().prepare_request(request)



from unittest import mock

import pytest

from requests_hardened import Manager, Config

DEFAULT_TIMEOUT = (1, 2)

TimeoutManager = Manager(
    Config(
        default_timeout=DEFAULT_TIMEOUT,
        # irrelevant:
        ip_filter_enable=False,
        ip_filter_allow_loopback_ips=True,
        never_redirect=False,
    )
)

NoTimeoutSetManager = TimeoutManager.clone()
NoTimeoutSetManager.config.default_timeout = None


@pytest.mark.parametrize(
    "_name, kwargs, expected_timeout",
    [
        ("Timeout not passed -> Should set default timeout", {}, DEFAULT_TIMEOUT),
        (
            "Timeout explicitly passed as 'None' -> Should set default timeout",
            {"timeout": None},
            DEFAULT_TIMEOUT,
        ),
        ("Timeout explicitly passed -> Should be untouched", {"timeout": 10}, 10),
        (
            "Timeout explicitly passed as tuple -> Should be untouched",
            {"timeout": (10, 20)},
            (10, 20),
        ),
    ],
)
@mock.patch("requests.sessions.Session.send")
def test_sets_default_timeout(
    mocked_send: mock.MagicMock, _name, kwargs, expected_timeout
):
    TimeoutManager.send_request("GET", "https://example.com", **kwargs)
    mocked_send.assert_called_once_with(
        mock.ANY,
        timeout=expected_timeout,
        allow_redirects=True,
        proxies={},
        stream=False,
        verify=True,
        cert=None,
    )


@pytest.mark.parametrize(
    "_name, kwargs, expected_timeout",
    [
        ("Timeout not passed -> Should set None", {}, None),
        (
            "Timeout explicitly passed as 'None' -> Should still be None",
            {"timeout": None},
            None,
        ),
        ("Timeout explicitly passed -> Should be untouched", {"timeout": 10}, 10),
    ],
)
@mock.patch("requests.sessions.Session.send")
def test_default_timeout_null(mocked_send: mock.MagicMock, _name, kwargs, expected_timeout):
    """Ensure when default timeout is set to None, then `timeout=` is set to None."""
    NoTimeoutSetManager.send_request("GET", "https://example.com", **kwargs)
    mocked_send.assert_called_once_with(
        mock.ANY,
        timeout=expected_timeout,
        allow_redirects=True,
        proxies={},
        stream=False,
        verify=True,
        cert=None,
    )




import ipaddress
import logging
import socket
from typing import Dict, Tuple, Union

import requests
from urllib3.util import Url, parse_url
from urllib3.util.connection import (  # type: ignore[attr-defined] # `allowed_gai_family` exists. # noqa: E501
    allowed_gai_family,
)

logger = logging.getLogger(__name__)


class InvalidIPAddress(requests.RequestException):
    pass


def get_ip_address(
    hostname: str, port: int, allow_loopback: bool
) -> Tuple[Union[ipaddress.IPv4Address, ipaddress.IPv6Address], int]:
    # Development safeguard, hostname shouldn't be surrounded by brackets
    # when calling this function.
    assert not hostname.startswith("[")

    addresses = socket.getaddrinfo(
        hostname, port, family=allowed_gai_family(), type=socket.SOCK_STREAM
    )

    af, socktype, proto, _canonname, socket_address = addresses[0]

    if af != socket.AF_INET and af != socket.AF_INET6:
        # This code shouldn't be reachable.
        raise ValueError(
            "Only AF_INET and AF_INET6 socket address families are supported"
        )
    port = socket_address[1]
    ip = ipaddress.ip_address(socket_address[0])

    # Python considers special IP ranges representing IPv4 addresses as being
    # private ranges, such as ::ffff:0:0/96, ::/128.
    # Because of that, we need to check the IPv4 address instead of the IPv6 one.
    # https://github.com/python/cpython/commit/ed391090cc8332406e6225d40877db6ff44a7104
    if ip.version == 6 and (ipv4 := ip.ipv4_mapped) is not None:
        ip = ipv4

    if allow_loopback and ip.is_loopback:
        return ip, port
    elif ip.is_private:
        logger.warning(
            "Forbidden IP address: %s for hostname %s",
            ip,
            hostname,
            # Extra arguments may clash with logger configuration
            # if it injects extra JSON fields, such as:
            # https://github.com/saleor/saleor/blob/5e7c57dad9e64b09477ebdcee53f0277359bc598/saleor/core/logging.py#L13
            # Because of that, we prefix the extra arguments in order to
            # reduce the chance of clashing with logging configs.
            extra={"dst_ip": ip, "dst_hostname": hostname},
        )
        raise InvalidIPAddress(ip)
    return ip, port


def filter_request(
    url: str, *, headers: Dict[str, Union[str, bytes, None]], allow_loopback: bool
) -> str:
    url = parse_url(url)
    port = url.port

    if not port:
        if url.scheme == "https":
            port = 443
        else:
            port = 80

    # IPv6 URL hostnames are embedded between "[" and "]"
    old_hostname = url.hostname
    if not old_hostname:
        raise ValueError("Invalid URL: missing hostname")
    if old_hostname.startswith("["):
        old_hostname = old_hostname.strip("[]")

    headers["Host"] = old_hostname
    ip_addr, port = get_ip_address(old_hostname, port, allow_loopback=allow_loopback)
    ip_addr_str = str(ip_addr) if ip_addr.version != 6 else f"[{ip_addr}]"

    url = Url(
        scheme=url.scheme,
        auth=url.auth,
        host=ip_addr_str,
        port=port,
        path=url.path,
        query=url.query,
        fragment=url.fragment,
    )
    return str(url)



from unittest import mock

import pytest

import requests

from requests_hardened import Manager, Config


UserAgentNotDefinedManager = Manager(
    Config(
        # irrelevant:
        never_redirect=True,
        ip_filter_enable=False,
        ip_filter_allow_loopback_ips=True,
        default_timeout=(1, 1),
    )
)
UserAgentDefinedManager = UserAgentNotDefinedManager.clone()
UserAgentDefinedManager.config.user_agent_override = None


@pytest.mark.parametrize("user_agent", ["user-agent", ""])
@mock.patch("requests.sessions.Session.send")
def test_user_agent_override_config(mocked_send: mock.MagicMock, user_agent):
    UserAgentDefinedManager.config.user_agent_override = user_agent
    UserAgentDefinedManager.send_request("GET", "https://example.com")

    prep_request = mocked_send.call_args[0][0]

    assert mocked_send.assert_called_once
    assert prep_request.headers.get("User-Agent") == user_agent


@mock.patch("requests.sessions.Session.send")
def test_user_agent_override_none(mocked_send: mock.MagicMock):
    UserAgentDefinedManager.config.user_agent_override = None
    UserAgentDefinedManager.send_request("GET", "https://example.com")

    prep_request = mocked_send.call_args[0][0]

    assert mocked_send.assert_called_once
    assert prep_request.headers.get("User-Agent") == f"python-requests/{requests.__version__}"


@mock.patch("requests.sessions.Session.send")
def test_user_agent_override_undefined(mocked_send: mock.MagicMock):
    UserAgentNotDefinedManager.send_request("GET", "https://example.com")

    prep_request = mocked_send.call_args[0][0]
    assert mocked_send.assert_called_once
    assert prep_request.headers.get("User-Agent") == f"python-requests/{requests.__version__}"



import ssl

from tests.http_test_servers import SNITLSHTTPTestServer
from tests.utils import get_remote_certificate


def test_dummy_sni_tls_server(tmp_path):
    """
    Tests that the SNI test server works as expected.

    The TLS server with SNI extension is expected to act as follows:
    1. domain.local returns the default certificate
    2. Invalid/unknown domains return domain.local's certificate (fallback)
    3. sni.local returns its own certificate.
    """

    srv = SNITLSHTTPTestServer(
        tmp_path,
        cert_identities=["domain.local"],
        cert_options={},
        additional_identities=["sni.local"],
    )

    with srv as sock_addr:
        ssl_ctx = ssl.create_default_context()
        ssl_ctx.load_verify_locations(srv.ca_bundle_path)

        # Retrieve the default TLS certificate (not an SNI callback one).
        cert_domain_local = get_remote_certificate(
            ssl_ctx, sock_addr, server_hostname="domain.local"
        )

        # Retrieve the certificate generated through SNI callback.
        cert_sni_local = get_remote_certificate(
            ssl_ctx, sock_addr, server_hostname="sni.local"
        )

        # Retrieve the TLS certificate of an nonexistent domain, it should return
        # the default certificate due to being unknown.
        ssl_ctx.check_hostname = False  # The hostname will mismatch thus ignore it
        cert_invalid_local = get_remote_certificate(
            ssl_ctx, sock_addr, server_hostname="invalid.local"
        )

    # domain.local and invalid.local should have returned the same certificate.
    assert cert_domain_local == cert_invalid_local

    # domain.local and sni.local should NOT have returned the same certificate.
    assert cert_domain_local["subjectAltName"] == (("DNS", "domain.local"),)
    assert cert_sni_local["subjectAltName"] == (("DNS", "sni.local"),)
